@@ -229,7 +229,7 @@ func TestExportFormsIIFE(t *testing.T) {
 		options: config.Options{
 			Mode:          config.ModeBundle,
 			OutputFormat:  config.FormatIIFE,
-			ModuleName:    "moduleName",
+			ModuleName:    []string{"moduleName"},
 			AbsOutputFile: "/out.js",
 		},
 	})
@@ -900,7 +900,7 @@ func TestRequireBadExtension(t *testing.T) {
 			Mode:          config.ModeBundle,
 			AbsOutputFile: "/out.js",
 		},
-		expectedScanLog: `/entry.js: error: File extension not supported: /test
+		expectedScanLog: `/entry.js: error: File could not be loaded: /test
 `,
 	})
 }
@@ -3169,6 +3169,25 @@ func TestInjectImportOrder(t *testing.T) {
 	})
 }
 
+func TestOutbase(t *testing.T) {
+	default_suite.expectBundled(t, bundled{
+		files: map[string]string{
+			"/a/b/c.js": `
+				console.log('c')
+			`,
+			"/a/b/d.js": `
+				console.log('d')
+			`,
+		},
+		entryPaths: []string{"/a/b/c.js", "/a/b/d.js"},
+		options: config.Options{
+			Mode:          config.ModeBundle,
+			AbsOutputDir:  "/out",
+			AbsOutputBase: "/",
+		},
+	})
+}
+
 func TestAvoidTDZNoBundle(t *testing.T) {
 	default_suite.expectBundled(t, bundled{
 		files: map[string]string{
@@ -3192,3 +3211,77 @@ func TestAvoidTDZNoBundle(t *testing.T) {
 		},
 	})
 }
+
+func TestProcessEnvNodeEnvWarning(t *testing.T) {
+	default_suite.expectBundled(t, bundled{
+		files: map[string]string{
+			"/entry.js": `
+				console.log(
+					process.env.NODE_ENV,
+					process.env.NODE_ENV,
+				)
+			`,
+		},
+		entryPaths: []string{"/entry.js"},
+		options: config.Options{
+			Mode:          config.ModeBundle,
+			AbsOutputFile: "/out.js",
+		},
+		expectedScanLog: `/entry.js: warning: Define "process.env.NODE_ENV" when bundling for the browser
+`,
+	})
+}
+
+func TestProcessEnvNodeEnvWarningNode(t *testing.T) {
+	default_suite.expectBundled(t, bundled{
+		files: map[string]string{
+			"/entry.js": `
+				console.log(process.env.NODE_ENV)
+			`,
+		},
+		entryPaths: []string{"/entry.js"},
+		options: config.Options{
+			Mode:          config.ModeBundle,
+			AbsOutputFile: "/out.js",
+			Platform:      config.PlatformNode,
+		},
+	})
+}
+
+func TestProcessEnvNodeEnvWarningDefine(t *testing.T) {
+	defines := config.ProcessDefines(map[string]config.DefineData{
+		"process.env.NODE_ENV": config.DefineData{
+			DefineFunc: func(loc logger.Loc, findSymbol config.FindSymbol) js_ast.E {
+				return &js_ast.ENull{}
+			},
+		},
+	})
+	default_suite.expectBundled(t, bundled{
+		files: map[string]string{
+			"/entry.js": `
+				console.log(process.env.NODE_ENV)
+			`,
+		},
+		entryPaths: []string{"/entry.js"},
+		options: config.Options{
+			Mode:          config.ModeBundle,
+			AbsOutputFile: "/out.js",
+			Defines:       &defines,
+		},
+	})
+}
+
+func TestProcessEnvNodeEnvWarningNoBundle(t *testing.T) {
+	default_suite.expectBundled(t, bundled{
+		files: map[string]string{
+			"/entry.js": `
+				console.log(process.env.NODE_ENV)
+			`,
+		},
+		entryPaths: []string{"/entry.js"},
+		options: config.Options{
+			Mode:          config.ModePassThrough,
+			AbsOutputFile: "/out.js",
+		},
+	})
+}
