@@ -1,7 +1,12 @@
 package config
 
 import (
+	"fmt"
+	"regexp"
+	"sync"
+
 	"github.com/evanw/esbuild/internal/compat"
+	"github.com/evanw/esbuild/internal/logger"
 )
 
 type LanguageTarget int8
@@ -24,26 +29,6 @@ const (
 )
 
 type StrictOptions struct {
-	// Loose:  "a ?? b" => "a != null ? a : b"
-	// Strict: "a ?? b" => "a !== null && a !== void 0 ? a : b"
-	//
-	// The disadvantage of strictness here is code bloat. The only observable
-	// difference between the two is when the left operand is the bizarre legacy
-	// value "document.all". This value is special-cased in the standard for
-	// legacy reasons such that "document.all != null" is false even though it's
-	// not "null" or "undefined".
-	NullishCoalescing bool
-
-	// Loose:  "a?.b" => "a == null ? void 0 : a.b"
-	// Strict: "a?.b" => "a === null || a === void 0 ? void 0 : a.b"
-	//
-	// The disadvantage of strictness here is code bloat. The only observable
-	// difference between the two is when the left operand is the bizarre legacy
-	// value "document.all". This value is special-cased in the standard for
-	// legacy reasons such that "document.all == null" is true even though it's
-	// not "null" or "undefined".
-	OptionalChaining bool
-
 	// Loose:  "class Foo { foo = 1 }" => "class Foo { constructor() { this.foo = 1; } }"
 	// Strict: "class Foo { foo = 1 }" => "class Foo { constructor() { __publicField(this, 'foo', 1); } }"
 	//
@@ -78,6 +63,7 @@ const (
 	LoaderFile
 	LoaderBinary
 	LoaderCSS
+	LoaderDefault
 )
 
 func (loader Loader) IsTypeScript() bool {
@@ -187,9 +173,10 @@ type Options struct {
 
 	OmitRuntimeForTests     bool
 	PreserveUnusedImportsTS bool
+	UseDefineForClassFields bool
 	AvoidTDZ                bool
+	ASCIIOnly               bool
 
-	Strict   StrictOptions
 	Defines  *ProcessedDefines
 	TS       TSOptions
 	JSX      JSXOptions
@@ -204,8 +191,9 @@ type Options struct {
 
 	AbsOutputFile     string
 	AbsOutputDir      string
+	AbsOutputBase     string
 	OutputExtensions  map[string]string
-	ModuleName        string
+	ModuleName        []string
 	TsConfigOverride  string
 	ExtensionToLoader map[string]Loader
 	OutputFormat      Format
@@ -213,6 +201,8 @@ type Options struct {
 	InjectAbsPaths    []string
 	InjectedFiles     []InjectedFile
 
+	Plugins []Plugin
+
 	// If present, metadata about the bundle is written as JSON here
 	AbsMetadataFile string
 
@@ -232,3 +222,113 @@ func (options *Options) OutputExtensionFor(key string) string {
 	}
 	return key
 }
+
+var filterMutex sync.Mutex
+var filterCache map[string]*regexp.Regexp
+
+func compileFilter(filter string) (result *regexp.Regexp) {
+	if filter == "" {
+		// Must provide a filter
+		return nil
+	}
+	ok := false
+
+	// Cache hit?
+	(func() {
+		filterMutex.Lock()
+		defer filterMutex.Unlock()
+		if filterCache != nil {
+			result, ok = filterCache[filter]
+		}
+	})()
+	if ok {
+		return
+	}
+
+	// Cache miss
+	result, err := regexp.Compile(filter)
+	if err != nil {
+		return nil
+	}
+
+	// Cache for next time
+	filterMutex.Lock()
+	defer filterMutex.Unlock()
+	if filterCache == nil {
+		filterCache = make(map[string]*regexp.Regexp)
+	}
+	filterCache[filter] = result
+	return
+}
+
+func CompileFilterForPlugin(pluginName string, kind string, filter string) (*regexp.Regexp, error) {
+	if filter == "" {
+		return nil, fmt.Errorf("[%s] %q is missing a filter", pluginName, kind)
+	}
+
+	result := compileFilter(filter)
+	if result == nil {
+		return nil, fmt.Errorf("[%s] %q filter is not a valid regular expression: %q", pluginName, kind, filter)
+	}
+
+	return result, nil
+}
+
+func PluginAppliesToPath(path logger.Path, filter *regexp.Regexp, namespace string) bool {
+	return (namespace == "" || path.Namespace == namespace) && filter.MatchString(path.Text)
+}
+
+////////////////////////////////////////////////////////////////////////////////
+// Plugin API
+
+type Plugin struct {
+	Name      string
+	OnResolve []OnResolve
+	OnLoad    []OnLoad
+}
+
+type OnResolve struct {
+	Name      string
+	Filter    *regexp.Regexp
+	Namespace string
+	Callback  func(OnResolveArgs) OnResolveResult
+}
+
+type OnResolveArgs struct {
+	Path       string
+	Importer   logger.Path
+	ResolveDir string
+}
+
+type OnResolveResult struct {
+	PluginName string
+
+	Path      logger.Path
+	External  bool
+	Namespace string
+
+	Msgs        []logger.Msg
+	ThrownError error
+}
+
+type OnLoad struct {
+	Name      string
+	Filter    *regexp.Regexp
+	Namespace string
+	Callback  func(OnLoadArgs) OnLoadResult
+}
+
+type OnLoadArgs struct {
+	Path logger.Path
+}
+
+type OnLoadResult struct {
+	PluginName string
+
+	Contents      *string
+	AbsResolveDir string
+	Loader        Loader
+
+	Msgs        []logger.Msg
+	ThrownError error
+}
