@@ -6,6 +6,7 @@ import (
 	"os"
 	"path/filepath"
 	"regexp"
+	"sort"
 	"strconv"
 	"strings"
 	"sync"
@@ -90,16 +91,21 @@ func validateLogLevel(value LogLevel) logger.LogLevel {
 	}
 }
 
-func validateStrict(value StrictOptions) config.StrictOptions {
-	return config.StrictOptions{
-		NullishCoalescing: value.NullishCoalescing,
-		OptionalChaining:  value.OptionalChaining,
-		ClassFields:       value.ClassFields,
+func validateASCIIOnly(value Charset) bool {
+	switch value {
+	case CharsetDefault, CharsetASCII:
+		return true
+	case CharsetUTF8:
+		return false
+	default:
+		panic("Invalid charset")
 	}
 }
 
 func validateLoader(value Loader) config.Loader {
 	switch value {
+	case LoaderNone:
+		return config.LoaderNone
 	case LoaderJS:
 		return config.LoaderJS
 	case LoaderJSX:
@@ -122,6 +128,8 @@ func validateLoader(value Loader) config.Loader {
 		return config.LoaderBinary
 	case LoaderCSS:
 		return config.LoaderCSS
+	case LoaderDefault:
+		return config.LoaderDefault
 	default:
 		panic("Invalid loader")
 	}
@@ -207,6 +215,22 @@ func validateFeatures(log logger.Log, target Target, engines []Engine) (compat.J
 	return compat.UnsupportedJSFeatures(constraints), compat.UnsupportedCSSFeatures(constraints)
 }
 
+func validateGlobalName(log logger.Log, text string) []string {
+	if text != "" {
+		source := logger.Source{
+			KeyPath:    logger.Path{Text: "(global path)"},
+			PrettyPath: "(global name)",
+			Contents:   text,
+		}
+
+		if result, ok := js_parser.ParseGlobalName(log, source); ok {
+			return result
+		}
+	}
+
+	return nil
+}
+
 func validateExternals(log logger.Log, fs fs.FS, paths []string) config.ExternalModules {
 	result := config.ExternalModules{
 		NodeModules: make(map[string]bool),
@@ -356,8 +380,8 @@ func validatePath(log logger.Log, fs fs.FS, relPath string) string {
 func validateOutputExtensions(log logger.Log, outExtensions map[string]string) map[string]string {
 	result := make(map[string]string)
 	for key, value := range outExtensions {
-		if key != ".js" {
-			log.AddError(nil, logger.Loc{}, fmt.Sprintf("Invalid output extension: %q (valid: .js)", key))
+		if key != ".js" && key != ".css" {
+			log.AddError(nil, logger.Loc{}, fmt.Sprintf("Invalid output extension: %q (valid: .css, .js)", key))
 		}
 		if !isValidExtension(value) {
 			log.AddError(nil, logger.Loc{}, fmt.Sprintf("Invalid output extension: %q", value))
@@ -367,20 +391,21 @@ func validateOutputExtensions(log logger.Log, outExtensions map[string]string) m
 	return result
 }
 
-func messagesOfKind(kind logger.MsgKind, msgs []logger.Msg) []Message {
+func convertMessagesToPublic(kind logger.MsgKind, msgs []logger.Msg) []Message {
 	var filtered []Message
 	for _, msg := range msgs {
 		if msg.Kind == kind {
 			var location *Location
+			loc := msg.Location
 
-			if msg.Location != nil {
-				loc := msg.Location
+			if loc != nil {
 				location = &Location{
-					File:     loc.File,
-					Line:     loc.Line,
-					Column:   loc.Column,
-					Length:   loc.Length,
-					LineText: loc.LineText,
+					File:      loc.File,
+					Namespace: loc.Namespace,
+					Line:      loc.Line,
+					Column:    loc.Column,
+					Length:    loc.Length,
+					LineText:  loc.LineText,
 				}
 			}
 
@@ -393,6 +418,35 @@ func messagesOfKind(kind logger.MsgKind, msgs []logger.Msg) []Message {
 	return filtered
 }
 
+func convertMessagesToInternal(msgs []logger.Msg, kind logger.MsgKind, messages []Message) []logger.Msg {
+	for _, message := range messages {
+		var location *logger.MsgLocation
+		loc := message.Location
+
+		if loc != nil {
+			namespace := loc.Namespace
+			if namespace == "" {
+				namespace = "file"
+			}
+			location = &logger.MsgLocation{
+				File:      loc.File,
+				Namespace: namespace,
+				Line:      loc.Line,
+				Column:    loc.Column,
+				Length:    loc.Length,
+				LineText:  loc.LineText,
+			}
+		}
+
+		msgs = append(msgs, logger.Msg{
+			Kind:     kind,
+			Text:     message.Text,
+			Location: location,
+		})
+	}
+	return msgs
+}
+
 ////////////////////////////////////////////////////////////////////////////////
 // Build API
 
@@ -410,27 +464,28 @@ func buildImpl(buildOpts BuildOptions) BuildResult {
 	options := config.Options{
 		UnsupportedJSFeatures:  jsFeatures,
 		UnsupportedCSSFeatures: cssFeatures,
-		Strict:                 validateStrict(buildOpts.Strict),
 		JSX: config.JSXOptions{
 			Factory:  validateJSX(log, buildOpts.JSXFactory, "factory"),
 			Fragment: validateJSX(log, buildOpts.JSXFragment, "fragment"),
 		},
-		Defines:           validateDefines(log, buildOpts.Defines, buildOpts.PureFunctions),
+		Defines:           validateDefines(log, buildOpts.Define, buildOpts.Pure),
 		Platform:          validatePlatform(buildOpts.Platform),
 		SourceMap:         validateSourceMap(buildOpts.Sourcemap),
 		MangleSyntax:      buildOpts.MinifySyntax,
 		RemoveWhitespace:  buildOpts.MinifyWhitespace,
 		MinifyIdentifiers: buildOpts.MinifyIdentifiers,
-		ModuleName:        buildOpts.GlobalName,
+		ASCIIOnly:         validateASCIIOnly(buildOpts.Charset),
+		ModuleName:        validateGlobalName(log, buildOpts.GlobalName),
 		CodeSplitting:     buildOpts.Splitting,
 		OutputFormat:      validateFormat(buildOpts.Format),
 		AbsOutputFile:     validatePath(log, realFS, buildOpts.Outfile),
 		AbsOutputDir:      validatePath(log, realFS, buildOpts.Outdir),
+		AbsOutputBase:     validatePath(log, realFS, buildOpts.Outbase),
 		AbsMetadataFile:   validatePath(log, realFS, buildOpts.Metafile),
 		OutputExtensions:  validateOutputExtensions(log, buildOpts.OutExtensions),
-		ExtensionToLoader: validateLoaders(log, buildOpts.Loaders),
+		ExtensionToLoader: validateLoaders(log, buildOpts.Loader),
 		ExtensionOrder:    validateResolveExtensions(log, buildOpts.ResolveExtensions),
-		ExternalModules:   validateExternals(log, realFS, buildOpts.Externals),
+		ExternalModules:   validateExternals(log, realFS, buildOpts.External),
 		TsConfigOverride:  validatePath(log, realFS, buildOpts.Tsconfig),
 		MainFields:        buildOpts.MainFields,
 		PublicPath:        buildOpts.PublicPath,
@@ -518,6 +573,8 @@ func buildImpl(buildOpts BuildOptions) BuildResult {
 		log.AddError(nil, logger.Loc{}, "Splitting currently only works with the \"esm\" format")
 	}
 
+	loadPlugins(&options, realFS, log, buildOpts.Plugins)
+
 	var outputFiles []OutputFile
 
 	// Stop now if there were errors
@@ -588,8 +645,8 @@ func buildImpl(buildOpts BuildOptions) BuildResult {
 
 	msgs := log.Done()
 	return BuildResult{
-		Errors:      messagesOfKind(logger.Error, msgs),
-		Warnings:    messagesOfKind(logger.Warning, msgs),
+		Errors:      convertMessagesToPublic(logger.Error, msgs),
+		Warnings:    convertMessagesToPublic(logger.Warning, msgs),
 		OutputFiles: outputFiles,
 	}
 }
@@ -605,25 +662,63 @@ func transformImpl(input string, transformOpts TransformOptions) TransformResult
 		LogLevel:      validateLogLevel(transformOpts.LogLevel),
 	})
 
+	// Settings from the user come first
+	preserveUnusedImportsTS := false
+	useDefineForClassFieldsTS := false
+	jsx := config.JSXOptions{
+		Factory:  validateJSX(log, transformOpts.JSXFactory, "factory"),
+		Fragment: validateJSX(log, transformOpts.JSXFragment, "fragment"),
+	}
+
+	// Settings from "tsconfig.json" override those
+	if transformOpts.TsconfigRaw != "" {
+		source := logger.Source{
+			KeyPath:    logger.Path{Text: "tsconfig.json"},
+			PrettyPath: "tsconfig.json",
+			Contents:   transformOpts.TsconfigRaw,
+		}
+		if result := resolver.ParseTSConfigJSON(log, source, nil); result != nil {
+			if len(result.JSXFactory) > 0 {
+				jsx.Factory = result.JSXFactory
+			}
+			if len(result.JSXFragmentFactory) > 0 {
+				jsx.Fragment = result.JSXFragmentFactory
+			}
+			if result.UseDefineForClassFields {
+				useDefineForClassFieldsTS = true
+			}
+			if result.PreserveImportsNotUsedAsValues {
+				preserveUnusedImportsTS = true
+			}
+		}
+	}
+
+	// Apply default values
+	if transformOpts.Sourcefile == "" {
+		transformOpts.Sourcefile = "<stdin>"
+	}
+	if transformOpts.Loader == LoaderNone {
+		transformOpts.Loader = LoaderJS
+	}
+
 	// Convert and validate the transformOpts
 	jsFeatures, cssFeatures := validateFeatures(log, transformOpts.Target, transformOpts.Engines)
 	options := config.Options{
-		UnsupportedJSFeatures:  jsFeatures,
-		UnsupportedCSSFeatures: cssFeatures,
-		Strict:                 validateStrict(transformOpts.Strict),
-		JSX: config.JSXOptions{
-			Factory:  validateJSX(log, transformOpts.JSXFactory, "factory"),
-			Fragment: validateJSX(log, transformOpts.JSXFragment, "fragment"),
-		},
-		Defines:           validateDefines(log, transformOpts.Defines, transformOpts.PureFunctions),
-		SourceMap:         validateSourceMap(transformOpts.Sourcemap),
-		OutputFormat:      validateFormat(transformOpts.Format),
-		ModuleName:        transformOpts.GlobalName,
-		MangleSyntax:      transformOpts.MinifySyntax,
-		RemoveWhitespace:  transformOpts.MinifyWhitespace,
-		MinifyIdentifiers: transformOpts.MinifyIdentifiers,
-		AbsOutputFile:     transformOpts.Sourcefile + "-out",
-		AvoidTDZ:          transformOpts.AvoidTDZ,
+		UnsupportedJSFeatures:   jsFeatures,
+		UnsupportedCSSFeatures:  cssFeatures,
+		JSX:                     jsx,
+		Defines:                 validateDefines(log, transformOpts.Define, transformOpts.Pure),
+		SourceMap:               validateSourceMap(transformOpts.Sourcemap),
+		OutputFormat:            validateFormat(transformOpts.Format),
+		ModuleName:              validateGlobalName(log, transformOpts.GlobalName),
+		MangleSyntax:            transformOpts.MinifySyntax,
+		RemoveWhitespace:        transformOpts.MinifyWhitespace,
+		MinifyIdentifiers:       transformOpts.MinifyIdentifiers,
+		ASCIIOnly:               validateASCIIOnly(transformOpts.Charset),
+		AbsOutputFile:           transformOpts.Sourcefile + "-out",
+		AvoidTDZ:                transformOpts.AvoidTDZ,
+		UseDefineForClassFields: useDefineForClassFieldsTS,
+		PreserveUnusedImportsTS: preserveUnusedImportsTS,
 		Stdin: &config.StdinInfo{
 			Loader:     validateLoader(transformOpts.Loader),
 			Contents:   input,
@@ -661,26 +756,162 @@ func transformImpl(input string, transformOpts TransformOptions) TransformResult
 	}
 
 	// Return the results
-	var js []byte
-	var jsSourceMap []byte
+	var code []byte
+	var sourceMap []byte
 
 	// Unpack the JavaScript file and the source map file
 	if len(results) == 1 {
-		js = results[0].Contents
+		code = results[0].Contents
 	} else if len(results) == 2 {
 		a, b := results[0], results[1]
 		if a.AbsPath == b.AbsPath+".map" {
-			jsSourceMap, js = a.Contents, b.Contents
+			sourceMap, code = a.Contents, b.Contents
 		} else if a.AbsPath+".map" == b.AbsPath {
-			js, jsSourceMap = a.Contents, b.Contents
+			code, sourceMap = a.Contents, b.Contents
 		}
 	}
 
 	msgs := log.Done()
 	return TransformResult{
-		Errors:      messagesOfKind(logger.Error, msgs),
-		Warnings:    messagesOfKind(logger.Warning, msgs),
-		JS:          js,
-		JSSourceMap: jsSourceMap,
+		Errors:   convertMessagesToPublic(logger.Error, msgs),
+		Warnings: convertMessagesToPublic(logger.Warning, msgs),
+		Code:     code,
+		Map:      sourceMap,
+	}
+}
+
+////////////////////////////////////////////////////////////////////////////////
+// Plugin API
+
+type pluginImpl struct {
+	log    logger.Log
+	fs     fs.FS
+	plugin config.Plugin
+}
+
+func (impl *pluginImpl) OnResolve(options OnResolveOptions, callback func(OnResolveArgs) (OnResolveResult, error)) {
+	filter, err := config.CompileFilterForPlugin(impl.plugin.Name, "OnResolve", options.Filter)
+	if filter == nil {
+		impl.log.AddError(nil, logger.Loc{}, err.Error())
+		return
+	}
+
+	impl.plugin.OnResolve = append(impl.plugin.OnResolve, config.OnResolve{
+		Name:      impl.plugin.Name,
+		Filter:    filter,
+		Namespace: options.Namespace,
+		Callback: func(args config.OnResolveArgs) (result config.OnResolveResult) {
+			response, err := callback(OnResolveArgs{
+				Path:       args.Path,
+				Importer:   args.Importer.Text,
+				Namespace:  args.Importer.Namespace,
+				ResolveDir: args.ResolveDir,
+			})
+			result.PluginName = response.PluginName
+
+			if err != nil {
+				result.ThrownError = err
+				return
+			}
+
+			if response.Namespace == "" {
+				response.Namespace = "file"
+			}
+			result.Path = logger.Path{Text: response.Path, Namespace: response.Namespace}
+			result.External = response.External
+
+			// Convert log messages
+			if len(response.Errors)+len(response.Warnings) > 0 {
+				msgs := make(sortableMsgs, 0, len(response.Errors)+len(response.Warnings))
+				msgs = convertMessagesToInternal(msgs, logger.Error, response.Errors)
+				msgs = convertMessagesToInternal(msgs, logger.Warning, response.Warnings)
+				sort.Sort(msgs)
+				result.Msgs = msgs
+			}
+			return
+		},
+	})
+}
+
+func (impl *pluginImpl) OnLoad(options OnLoadOptions, callback func(OnLoadArgs) (OnLoadResult, error)) {
+	filter, err := config.CompileFilterForPlugin(impl.plugin.Name, "OnLoad", options.Filter)
+	if filter == nil {
+		impl.log.AddError(nil, logger.Loc{}, err.Error())
+		return
+	}
+
+	impl.plugin.OnLoad = append(impl.plugin.OnLoad, config.OnLoad{
+		Filter:    filter,
+		Namespace: options.Namespace,
+		Callback: func(args config.OnLoadArgs) (result config.OnLoadResult) {
+			response, err := callback(OnLoadArgs{
+				Path:      args.Path.Text,
+				Namespace: args.Path.Namespace,
+			})
+			result.PluginName = response.PluginName
+
+			if err != nil {
+				result.ThrownError = err
+				return
+			}
+
+			result.Contents = response.Contents
+			result.Loader = validateLoader(response.Loader)
+			if absPath := validatePath(impl.log, impl.fs, response.ResolveDir); absPath != "" {
+				result.AbsResolveDir = absPath
+			}
+
+			// Convert log messages
+			if len(response.Errors)+len(response.Warnings) > 0 {
+				msgs := make(sortableMsgs, 0, len(response.Errors)+len(response.Warnings))
+				msgs = convertMessagesToInternal(msgs, logger.Error, response.Errors)
+				msgs = convertMessagesToInternal(msgs, logger.Warning, response.Warnings)
+				sort.Sort(msgs)
+				result.Msgs = msgs
+			}
+			return
+		},
+	})
+}
+
+// This type is just so we can use Go's native sort function
+type sortableMsgs []logger.Msg
+
+func (a sortableMsgs) Len() int          { return len(a) }
+func (a sortableMsgs) Swap(i int, j int) { a[i], a[j] = a[j], a[i] }
+
+func (a sortableMsgs) Less(i int, j int) bool {
+	ai := a[i].Location
+	aj := a[j].Location
+	if ai == nil || aj == nil {
+		return ai == nil && aj != nil
+	}
+	if ai.File != aj.File {
+		return ai.File < aj.File
+	}
+	if ai.Line != aj.Line {
+		return ai.Line < aj.Line
+	}
+	if ai.Column != aj.Column {
+		return ai.Column < aj.Column
+	}
+	return a[i].Text < a[j].Text
+}
+
+func loadPlugins(options *config.Options, fs fs.FS, log logger.Log, plugins []Plugin) {
+	for i, item := range plugins {
+		if item.Name == "" {
+			log.AddError(nil, logger.Loc{}, fmt.Sprintf("Plugin at index %d is missing a name", i))
+			continue
+		}
+
+		impl := &pluginImpl{
+			fs:     fs,
+			log:    log,
+			plugin: config.Plugin{Name: item.Name},
+		}
+
+		item.Setup(impl)
+		options.Plugins = append(options.Plugins, impl.plugin)
 	}
 }
